/***************************************************************************/
/*                                                                         */
/* FILE NAME : ga.h                                                        */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

/* 
   Definition
*/

#define MAX_ND 100
#define MAX_POP 1000
#define aa 16807
#define mm 2147483647
#define qq 127773
#define rr 2836  


/***************************************************************************/
/*                                                                         */
/* FILE NAME : ga-for-gp.h                                                 */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

/*
  Program for graph partitioning using GA's, written by Dr. Ching-Tai Chiu 
*/

#include "ga.h"
#include <stdio.h>

extern void generate_initial_set();
extern void generate_graph();
extern double calc_pop_fitness(), calc_fitness();
extern void reproduction();
extern void rescale();
int evol, spop, besti;
double zs[MAX_POP], w[MAX_ND][MAX_ND];
float p[MAX_POP];
int nodecount;
int bstr[MAX_POP][MAX_ND], bestsol[MAX_ND];
FILE *fp;

DebugPrint(str)
char *str;
{ printf("\n %s \n",str);
  return;
}

main(argc, argv)
int argc;
char *argv[];
{
  int i, j, worsti, count0, count1;
  double avr_fit, zmax;
  long int clock;
  FILE *mpx, *mpy;
  char mfx[30], mfy[30];

  if (argc==1)
    {
      printf("Usage:\n");
      printf("nn num_nodes population epochs outputfile\n");
      exit(0);
    }
  sscanf(argv[1],"%d", &nodecount);
/*  nodecount = atoi(argv[1]);  */
  sscanf(argv[2], "%d", &spop);
/*  spop = atoi(argv[2]);  */
  sscanf(argv[3], "%d", &evol);
/*  evol = atoi(argv[3]);  */

  if ((fp = fopen(argv[4], "w+t")) == NULL)
    {
      printf("Could not open output file\n");
      exit(0);
    }

  DebugPrint("I opened the output file ");

  rewind(fp);

  strcpy(mfx, argv[4]);
  strcpy(mfy, argv[4]);
  strcat(mfx, "x.mat");
  strcat(mfy, "y.mat");
  mpx = fopen(mfx, "w+t");
  mpy = fopen(mfy, "w+t");

  generate_graph();

  DebugPrint(" I generated GRAPH "); 

  clock = time(0);
  generate_initial_set();
  avr_fit = calc_pop_fitness();
  rescale();

  for (i=0; i<spop; i++) {
    fprintf(fp, "\n%d    %6.2f  %6.3f --- ", i, zs[i], p[i]);
    for (j=0; j<nodecount; j++) fprintf(fp, "%d", bstr[i][j]);
  }

  fprintf(fp, "\nAverage fitness = %8.4f    Min[%d]=%6.3f\n", 
	  avr_fit, besti, zs[besti]);

  fprintf(mpx, "%d\n", 0);
  fprintf(mpy, "%5.0f\n", zs[besti]);

  for (i=0; i<evol; i++) {
   
    DebugPrint("EVOL");

    reproduction();

    DebugPrint("EVOL1");

/*    zmax = zs[0];
    worsti = 0;
    for (j=0; j<spop; j++) 
      if (zmax < zs[j]) {
	worsti = j;
	zmax = zs[j];
      }
    for (j=0; j<nodecount; j++)	bstr[worsti][j] = bestsol[j];
*/
    avr_fit = calc_pop_fitness();

    DebugPrint("EVOL2");

    rescale();

    DebugPrint("EVOL3");

    printf("\nGeneration %4d: average fitness = %8.4f    Min[%d]=%6.3f\n", 
	   i+1, avr_fit, besti, zs[besti]);

    fprintf(mpx, "%d\n", i+1);
    fprintf(mpy, "%5.0f\n", zs[besti]);

    if ((i % 10) == 0)
      fprintf(fp, "\n%5d: Average fitness = %8.4f  Min[%d]=%6.3f", 
	      i+1, avr_fit, besti, zs[besti]);   


    if (drand48()<0.02 || (avr_fit - zs[besti]) < 0.005) {

      printf("\nMutation!!");
      mutation();
      avr_fit = calc_pop_fitness();
      rescale();
      i++;
      fprintf(fp, "\n%5d: Average fitness = %8.4f  Min[%d]=%6.3f", 
	      i+1, avr_fit, besti, zs[besti]);
      fprintf(mpx, "%d\n", i+1);
      fprintf(mpy, "%5.0f\n", zs[besti]);

/*      zmax = zs[0];
      worsti = 0;
      for (j=0; j<spop; j++) 
	if (zmax < zs[j]) {
	  worsti = j;
	  zmax = zs[j];
	}
      for (j=0; j<nodecount; j++) bstr[worsti][j] = bestsol[j];  */


    }
  }

  clock = time(0) - clock;
  for (i=0; i<spop; i++) {
    fprintf(fp, "\n%d    %6.2f  %6.3f --- ", i, zs[i], p[i]);
    for (j=0; j<nodecount; j++) fprintf(fp, "%d", bstr[i][j]);
  }

  fprintf(fp, "\nNodes = %d\nPopulation = %d\nGenerations = %d",
	  nodecount, spop, evol);

  fprintf(fp, "\n\nBest cost = %8.2f -> ", calc_fitness(bstr[besti]));
  count0 = count1 = 0;
  for (i=0; i<nodecount; i++) 
    {
      fprintf(fp, "%d", bstr[besti][i]);
      if (bstr[besti][i] == 0)
	count0++;
      else
	count1++;
    }
  fprintf(fp, "\n\n%d nodes in 0 set\n%d nodes in 1 set\n", count0, count1);

  fprintf(fp, "\nTime = %d\n", clock);

  fclose(mpx);
  fclose(mpy);
  fclose(fp);
}


/***************************************************************************/
/*                                                                         */
/* FILE NAME : ga_operator.c                                               */
/*                                                                         */
/* GA operators                                                            */
/*                                                                         */
/***************************************************************************/

#include "ga.h"
#include <stdio.h>

extern double zs[MAX_POP];
extern float p[MAX_POP];
extern int spop, nodecount;
extern int bstr[MAX_POP][MAX_ND];
extern FILE *fp;
extern int rnd();
extern double calc_fitness();

void select_maters(copy)
int *copy;
{
  int j,k, nassign;
  float fraction[MAX_POP];

  printf("before first while \n");

  j = k = 0;

  while ((j < spop)&&(j<MAX_POP)) 
  {
     nassign = p[j];
     fraction[j] = p[j] - nassign;

     printf("before inside first while \n");

     while (nassign > 0) 
     {
       nassign--;
       *(copy+k) = j;
       k++;
     }/* end of while */

     j++;

  }/* end of while */
     
  printf("After first while \n");

  j = 0;

  while(k < spop)
    {
      if (j >= spop) j=0;
/*      printf("fraction[%d] = %6.2f\n", j, fraction[j]); */
      if (fraction[j] > 0) 
	if ( preflip( fraction[j] ) == 1 ) {
	  *(copy+k) = j;
	  fraction[j]--;
	  k++;
	}
      j++;
    }

  printf("After second while \n");

}


int select_mate(copyp, remp)
int *copyp, *remp;
{
  int ilast, pick, tempi;

  ilast = --(*remp);
  pick = rnd(0, ilast);
  tempi = *(copyp+pick);
  *(copyp+pick) = *(copyp+ilast);
  *(copyp+ilast) = tempi;
  return (tempi);
}


adjust_bit(xp)
int *xp;
{
  int i, ln = nodecount, count0, count1;

  i = rnd(0, nodecount-1);
  count0 = count1 = 0;

  while (ln > 0) 
    {
      if (i==nodecount) i=0;
      if(*(xp+i) == 0) 
	{
	  if (count0 < nodecount/2)
	    count0++;
	  else 
	    {
	      *(xp+i) = 1;
	      count1++;
	    }
	}
      else
	{
	  if (count1 < nodecount/2)
	    count1++;
	  else
	    {
	      *(xp+i) = 0;
	      count0++;
	    }
	}
      i++;
      ln--;
    }
}
	

void xover(dadp, momp, xdp, xmp)
int dadp, momp;
int *xdp, *xmp;
{
  int i, startp, lnth, tempb;
  int *temp;
  double zsd, zsm, ztemp;
  
  if (zs[dadp] < zs[momp]) 
    {
      tempb = dadp;
      dadp = momp;
      momp = tempb;
    }

  for (i=0; i<nodecount; i++)
    {
      *(xdp+i) = bstr[dadp][i];
      *(xmp+i) = bstr[momp][i];
    }
  
  if (dadp == momp) return; /* No crossover needed */

  startp = rnd(0, nodecount-1);
  lnth = rnd(1, nodecount-1);

  if ( (startp+lnth) > nodecount )
    {
      startp = startp + lnth - nodecount;
      lnth = startp + nodecount - lnth;
    }

  for (i=startp; i<lnth; i++)
    {
      tempb = *(xdp+i);
      *(xdp+i) = *(xmp+i);
      *(xmp+i) = tempb;
    }

  adjust_bit(xdp);
  adjust_bit(xmp);

  zsd = calc_fitness(xdp);
  zsm = calc_fitness(xmp);

  if (zsd < zsm)
    {
      temp = xdp;
      xdp = xmp;
      xmp = temp;
      ztemp = zsd;
      zsd = zsm;
      zsm = ztemp;
    } 

/* zsd > zsm > zs[dadp] > zs[momp] */
  if (zsm >= zs[dadp]) 
    {
      for (i=0; i<nodecount; i++) *(xdp+i) = bstr[dadp][i];
      for (i=0; i<nodecount; i++) *(xmp+i) = bstr[momp][i];
    }  
/* (zsd, zs[dadp]) >= zsm >= zs[momp] */
  else if (zsm >= zs[momp]) 
    for (i=0; i<nodecount; i++) *(xdp+i) = bstr[momp][i];  
/* zsd > zs[dadp] > zs[momp] > zsm */
  else if (zsd > zs[dadp])  
    {
      for (i=0; i<nodecount; i++) *(xdp+i) = bstr[momp][i];
      if (zsm==0) for (i=0; i<nodecount; i++) *(xmp+i) = bstr[dadp][i];
    } 
/* zs[dadp] > zsd > zs[momp] > zsm */
  else if (zsd > zs[momp]) 
    {
      if (zsm==0) for (i=0; i<nodecount; i++) *(xmp+i) = *(xdp+i);
      for (i=0; i<nodecount; i++) *(xdp+i) = bstr[momp][i];

    } 
/* zs[dadp] > zs[momp] > zsd > zsm */
  else  
    {
      if (zsm==0) {
	for (i=0; i<nodecount; i++) *(xmp+i) = bstr[momp][i];
	if (zsd == 0) for (i=0; i<nodecount; i++) *(xdp+i) = bstr[dadp][i];
      }
    }  


/*  fprintf(fp,"\nxdp = ");
  for (i=0; i<nodecount; i++)
      fprintf(fp, "%d", *(xdp+i));
  fprintf(fp,"\nxmp = ");
  for (i=0; i<nodecount; i++)
      fprintf(fp, "%d", *(xmp+i));   */

}


void reproduction()
{
  int copy[MAX_POP],i, j, ngene[MAX_POP][MAX_ND];
  int remp, dad, mom, xstrd[MAX_ND], xstrm[MAX_ND];

  printf("\n Inside REPRODUCTION : before select_maters \n");

  select_maters(copy);

  printf("\n Inside REPRODUCTION \n");

/*  for (i=0; i<spop; i++) fprintf(fp, "copy[%d] = %d\n", i, copy[i]); */

  remp = spop;

  for (i=0; i<spop; i += 2)
    {
      dad = select_mate(copy, &remp);
      mom = select_mate(copy, &remp);
/*      fprintf(fp, "dad = %3d  mom = %3d\n", dad, mom);   */

      printf("before xover\n");

      xover(dad, mom, xstrd, xstrm);

      printf("after xover\n");

      for (j=0; j<nodecount; j++)
	{
	  ngene[i][j] = xstrd[j];
	  ngene[i+1][j] = xstrm[j];
	}
    }

  for (i=0; i<spop; i++) 
    for (j=0; j<nodecount; j++) bstr[i][j] = ngene[i][j];

  printf("\n End of REPRODUCTION \n");

/*  for (i=0; i<spop; i++) fprintf(fp, "copy[%d] = %d\n", i, copy[i]); */
}

void select_mutant(remp)
int *remp;
{
  int ilast, pick, tempi, i, count0, count1;
  int mutant[MAX_ND], startp, lnth;

  ilast = --(*remp);
  pick = rnd(0, ilast);
  for (i=0; i<nodecount; i++) mutant[i] = bstr[pick][i];
  for (i=0; i<nodecount; i++) bstr[pick][i] = bstr[ilast][i];

  startp = rnd(0, nodecount-1);
  lnth = rnd(1, nodecount-1)/4;
  if (lnth == 0) lnth = 1; 
  count0 = count1 = 0;
  i = startp;
  while (count0 < lnth || count1 < lnth)
    {
      if (mutant[i] == 0 && count0<lnth)
	{
	  mutant[i] = 1;
	  count0++;
	}
      if (mutant[i] == 1 && count1<lnth)
	{
	  mutant[i] = 0;
	  count1++;
	}
      if ( (++i) == nodecount ) i=0;
    }

  for (i=0; i<nodecount; i++) bstr[ilast][i] = mutant[i];      
}


void mutation()
{
  int remp=spop, i;

  for (i=0; i<(int) (spop/4); i++) select_mutant(&remp);
}



/***************************************************************************/
/*                                                                         */
/* FILE NAME : inp_gen.c                                                   */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

/*
  Randomly generate a set of binary string
*/

#include "ga.h"
#include <stdio.h>

extern int preflip();
extern int evol, spop, nodecount;
extern double w[MAX_ND][MAX_ND];
extern int bstr[MAX_POP][MAX_ND];
extern long int fpseed;
extern double drand48(), calc_fitness();
extern FILE *fp;

void generate_graph()
{
  int 		i,j;
  double u;


  printf("\nStep 1: Generate graph randomly");
  printf("\n\nPlease enter number of nodes (10-100):");
  scanf("%d",&nodecount);


  for(i=0; i< nodecount; i++){
    w[i][i]=0;
    for(j=i; j < nodecount; j++){
      if (i!=j)
	if ( drand48() > 0.5 )
	  w[i][j]=(int)(10.0* drand48() -0.5)+1;
	else 
	  w[i][j]=0;
      w[i][i] += w[i][j];
    }
    for(j=0; j< i; j++){
      w[i][j]=w[j][i];
      w[i][i] += w[i][j];
    }
  }

/*  for (i=0; i<nodecount; i++) {
    for (j=0; j<nodecount; j++)
      fprintf(fp, "%3.0f ", w[i][j]);
    fprintf(fp, "\n");  
  }  
*/
}


void generate_initial_set()
{
  int i, j, k, count0, count1, tempb;

/*
  printf("\n\nStep 2: Generate initial solutions");
  printf("\n\nPlease enter the size of population (10-1000):");
  scanf("%d", &spop);

  printf("\n\nStep 3: Start evolution");
  printf("\n\nHow many generations do you want to simulate?");
  scanf("%d", &evol);
*/

  for (i=0; i< MAX_POP; i++)
    for (j=0; j < MAX_ND; j++) bstr[i][j] = -1;

  for (i=0; i<spop; i++) 
    {
      count0 = count1 = 0;
      j = 0;
      while (j<nodecount) 
	{
	  if ((tempb = preflip(0.5)) ==0) 
	    count0++;
	  else
	    count1++;
	  bstr[i][j] = tempb;
	  k = ++j;
	  if (count0== (nodecount/2)) 
	    for (j=k; j<nodecount; j++) bstr[i][j] = 1;
	  else if (count1== (nodecount/2) ) 
	    for (j=k; j<nodecount; j++) bstr[i][j] = 0;
	}
    }
}


/***************************************************************************/
/*                                                                         */
/* FILE NAME : rnd_ut.h                                                    */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

#include <stdio.h>
#include "ga.h"

extern double drand48();
long int fpseed = 150;


int preflip(prob)
float prob;
{
  long int lo, hi, test;
  float u;

  hi = fpseed / qq;
  lo = fpseed % (long int) qq;
  test = aa*lo - rr*hi;

  if (test > 0) 
    fpseed = test;
  else
    fpseed = test + mm;

  u =  srand(fpseed);
  u = u / mm;
  if (u < prob)
    return 0;
  else
    return 1;
}


int rnd(low, high)
int low, high;
{
  double u;
  int n;

  u = drand48();  
  n = u*(high-low+1) + low;
  if (n>high) n = high;
  return (n);
}


/***************************************************************************/
/*                                                                         */
/* FILE NAME : fitness.c                                                   */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

#include "ga.h"

extern int nodecount, spop, besti;
extern double zs[MAX_POP];
extern int bstr[MAX_POP][MAX_ND], bestsol[MAX_ND];
extern double w[MAX_ND][MAX_ND];
extern float p[MAX_POP];

void rescale()
{
  double zmin, zmax;
  float maxprob, minprob;
  int i;

  maxprob = 1.2;
  minprob = 0.8;
  zmin = zmax = zs[0];
  besti = 0;
  for (i=1; i<spop; i++)
    {
      if (zs[i] > zmax) zmax = zs[i];
      if (zs[i] < zmin) {
	zmin = zs[i];
	besti = i;
      }
    }
  for (i=0; i<nodecount; i++) bestsol[i] = bstr[besti][i];
  
  for (i=0; i<spop; i++)
    p[i] = maxprob - ( (zs[i]-zmin)/(zmax-zmin) ) * (maxprob - minprob);
}

double 
calc_fitness(xp)
int *xp;
{
  int j, k;
  double cons = 0;
  
  for (j = 0; j < nodecount; j++) {
    for (k = 0; k < nodecount; k++) {
      if (j != k)
	cons = cons - (w[j][k] * (*(xp+j)) * (*(xp+k)));
    }
    cons += w[j][j] * (*(xp+j));
  }

  return cons;
}

double calc_pop_fitness()
{
  int i, j, count0, count1;
  double avr = 0;

  for (i=0; i<spop; i++)
    {
      count0 = count1 =0;
      for (j=0; j<nodecount; j++) {
	if ( bstr[i][j] == 0)
	  count0++;
	else
	  count1++;
      }

      zs[i] = calc_fitness(bstr[i]);
/*  + (count0-count1)*(count0-count1);   */
      avr += zs[i];
    }

  return (avr/spop);
}

